
// Driver_3_05_09

`include "dbgflags.h"
`include "stdbus.h"
`include "memory.h"
`include "trace.h"


module Driver (
        Strobe,
        RWSelect,
        Address,
        Data,
        Ready,
        Reset,
        Clk );

output          Strobe;
output          RWSelect;
output  [`ADDR] Address;
inout   [`DATA] Data;
input           Ready;
input           Reset;
input           Clk;

reg             BusStrobe;
reg     [`ADDR] MAR;
reg     [`DATA] MDR;
reg             BusOE;

wire            Strobe   = BusStrobe;
reg             RWSelect;
wire    [`ADDR] Address  = MAR;
wire    [`DATA] Data     = BusOE ? MDR : `DATAWIDTH'bz;


// Trace declarations
reg [`TRACE] TMem[0:`MAXTRACESIZE-1];
reg [`TRACE] Ttmp;
reg [`TACCS] Tacc;
reg [`TADDR] Taddr;

// Read signature to verify correct cache behavior
reg [`DATA] ReadSignature, WriteSignature;

integer i, n, ndr, ndw, nif;
reg [`TADDR] minifaddr, maxifaddr;


// Cache access statistics
integer nreadhits, nreadmiss;
integer nwritehits, nwritemiss,nwriteback;

// Timing statistics declarations
integer readtime;
integer writetime;


task Read;
input   [`ADDR] ReadAddr;
integer         flag;
begin
   // Initiate memory read request
   if (`vbs) $display($time, " driver> Read Strobe (%h,%h)", ReadAddr, 'bx);
   @ (posedge Clk) begin
      MAR       = ReadAddr;
      BusStrobe = `ENABLE;
      RWSelect  = `READ;
      BusOE     = `DISABLE;
   end
   @ (posedge Clk) begin
      if (`dbg) $display($time, " driver> Strobe disable. (Ready %b)", Ready);
      BusStrobe = `DISABLE;
   end

   // Wait for memory response
   //!!
   //!! WARNING: Wrapping the Ready wait statement into the while loop
   //!!          is a kludge! The Ready signal (continuously assigned
   //!!          in control.v) may jump from 0 to x during reevaluation.
   //!!          As a result, the wait statement is triggered erroneously!
   //!!
   flag = 1;
   while (flag) begin
      @ (posedge Ready) begin
         if (Ready == 1) begin
            if (`dbg) $display($time, " driver> Reading ... (Ready %b)", Ready);
            #(`CLOCKPERIOD/2) MDR = Data;
            ReadSignature = ReadSignature + Data;
            if (`vbs) $display($time, " driver> Read  (%h,%h)", ReadAddr, MDR);
            flag = 0;
            // Wait until end of clock cycle while in STATE_READDATA
            @ (posedge Clk)
              if (`dbg) $display($time, " driver> Read Done!");
         end
      end
   end
end
endtask

task Write;
input   [`ADDR] WriteAddr;
input   [`DATA] WriteData;
integer         flag;
begin
   // Initiate memory write request
   if (`vbs) $display($time, " driver> Write Strobe (%h,%h)", WriteAddr, WriteData);
   @ (posedge Clk) begin
      MAR       = WriteAddr;
      MDR       = WriteData;
      BusStrobe = `ENABLE;
      RWSelect  = `WRITE;
      BusOE     = `ENABLE;
   end
   @ (posedge Clk) begin
      if (`dbg) $display($time, " driver> Strobe disable.");
      BusStrobe = `DISABLE;
   end

   // Wait for memory response
   //!!
   //!! WARNING: Wrapping the Ready wait statement into the while loop
   //!!          is a kludge! The Ready signal (continuously assigned
   //!!          in control.v) may jump from 0 to x during reevaluation.
   //!!          As a result, the wait statement is triggered erroneously!
   //!!
   flag = 1;
   while (flag) begin
      @ (posedge Ready) begin
         if (Ready == 1) begin
            // Wait one more clock cycle while in STATE_WRITEDATA
            #1 @(posedge Clk) begin
               BusOE = `DISABLE;
               if (`dbg) $display($time, " driver> Write Done!");
               flag = 0;
            end
         end
      end
   end
end
endtask


task trace_reset;
begin
   $display(" trace reset ...");
   n   = 0;
   ndr = 0;
   ndw = 0;
   nif = 0;
   ReadSignature = 0;
   for (i=0; i<`MAXTRACESIZE; i=i+1)
      TMem[i] = `TRACEWIDTH'bx;
end
endtask

task trace_stat;
begin
   $display(" trace statistics ...");
   n         = 0;
   minifaddr = 'hFFFFFFFF;
   maxifaddr = 0;
   Ttmp      = TMem[n];
   Tacc      = Ttmp[`TACCS];
   Taddr     = Ttmp[`TADDR];
   /* Force address to word boundary
      Not all addresses in trace are word aligned
   */
   Taddr = {Taddr[`ADDRWIDTH-1:2],2'b0};

   while (Tacc !== 4'bx) begin
      if (Tacc === `DR) begin
         ndr = ndr + 1;
         // Initialize read data address
         //Main.HashMem.writemem(Taddr, 0);
         Main.HashMem.writemem(Taddr, n);
         Main.HashMem.writemem({Taddr[`ADDRWIDTH-1:3],!Taddr[2],2'b0},ndr);
      end
      else if (Tacc === `DW)begin
         ndw = ndw + 1;
         Main.HashMem.writemem({Taddr[`ADDRWIDTH-1:3],!Taddr[2], 2'b0},ndw);
      end
      else if (Tacc === `IF) begin
         nif = nif + 1;
         if (minifaddr > Taddr)
            minifaddr = Taddr;
         if (maxifaddr < Taddr)
            maxifaddr = Taddr;
         // Initialize instruction address
         Main.HashMem.writemem(Taddr, n);
         Main.HashMem.writemem({Taddr[`ADDRWIDTH-1:3],!Taddr[2],2'b0},nif);
      end
      else begin
         $display(" Illegal trace access code: %d", Tacc);
         $finish;
      end
      n     = n + 1;
      Ttmp  = TMem[n];
      Tacc  = Ttmp[`TACCS];
      Taddr = Ttmp[`TADDR];
      /* Force address to word boundary
         Not all addresses in trace are word aligned
      */
       Taddr = {Taddr[`ADDRWIDTH-1:2],2'b0};
   end
   $display(" number of memory accesses              %d", n);
   $display(" number of data reads                   %d", ndr);
   $display(" number of data writes                  %d", ndw);
   $display(" number of instruction fetches          %d", nif);
   $display(" minimum instruction address             %d", minifaddr);
   $display(" maximum instruction address             %d", maxifaddr);
end
endtask

task compute_write_signature; // for verifying correct cache operation
//The signatures of all cache organizations should match for a given trace file
begin
   
   WriteSignature = 0;
   n         = 0;
   
   Ttmp      = TMem[n];
   Tacc      = Ttmp[`TACCS];
   Taddr     = Ttmp[`TADDR];
   
   /* Force address to word boundary
         Not all addresses in trace are word aligned
      */
   Taddr = {Taddr[`ADDRWIDTH-1:2],2'b0};
   
   while (Tacc !== 4'bx) begin
      if (Tacc === `DW) begin
         // read final value of memory location and apply to signature
         WriteSignature = WriteSignature + Main.HashMem.readmem(Taddr);
      end
      n     = n + 1;
      Ttmp  = TMem[n];
      Tacc  = Ttmp[`TACCS];
      Taddr = Ttmp[`TADDR];
      /* Force address to word boundary
        Not all addresses in trace are word aligned
      */
   Taddr = {Taddr[`ADDRWIDTH-1:2],2'b0};
   end
end
endtask

task run_trace;
integer i, begt;
begin
   $display(" ");
   $display(" run trace (%d accesses) ...", n);

   for(i=0; i<n; i=i+1) begin
      Ttmp  = TMem[i];
      Tacc  = Ttmp[`TACCS];
      Taddr = Ttmp[`TADDR];
      /* Force to word address
         Some addresses in trace file are not word aligned
     */
      Taddr = {Taddr[`ADDRWIDTH-1:2],2'b0};
      if (`vbs) $display(" %d> acc %d   addr %h", i, Tacc, Taddr);

      if (Tacc == `IF || Tacc == `DR) begin
         begt = $time;
         Read(Taddr);
         readtime = readtime + $time - begt;
      end
      else begin        // data write `DW
         begt = $time;
         Write(Taddr, i);
         writetime = writetime + $time - begt;
      end

   end
end
endtask


initial begin
   BusStrobe = `DISABLE;
   BusOE     = `DISABLE;
end

initial begin : run_memory_trace
   trace_reset;
   $readmemh(`TRACEFILE, TMem);
   // Delay trace_stat to make sure, initializations take effect.
   #10 trace_stat;

   nreadhits  = 0;
   nreadmiss  = 0;
   nwritehits = 0;
   nwritemiss = 0;
   nwriteback = 0;
   
   readtime   = 0;
   writetime  = 0;

   run_trace;
   compute_write_signature;

   $display(" ");
   $display(" trace file                %s", `TRACEFILE);
   $display(" ");
   $display(" read wait cycles           %d", `READ_WAITCYCLES);
   $display(" write wait cycles          %d", `WRITE_WAITCYCLES);
   $display(" ");
   $display(" cache read hits            %d", nreadhits);
   $display(" cache read misses          %d", nreadmiss);
   $display(" cache write hits           %d", nwritehits);
   $display(" cache write misses         %d", nwritemiss);
   $display(" cache write-backs          %d", nwriteback);
   $display(" ");
   $display(" memory read time           %d", readtime);
   $display(" memory write time          %d", writetime);
   $display(" Read signature              %d", ReadSignature);
   $display(" Write Signature             %d", WriteSignature);
   $finish;
end


endmodule /* Driver */
