
`include "dbgflags.h"
`include "memory.h"
`include "trace.h"
`include "stdbus.h"


// Large Memory simulated with hash table.
// Hash strategy: Open addressing with double hashing
//
// HASHSIZE must be a power of two for double hashing to work!
//`define HASHSIZE       4096
//`define HASHSIZE       16384
`define HASHSIZE	 64536


module HashMem (Strobe, RWSelect, Addr, Data, Clk);
input           Strobe;
input           RWSelect;
input   [`ADDR] Addr;
inout   [`DATA] Data;
input           Clk;

reg     [`DATA] SReg;
wire    [`DATA] Data = RWSelect == `READ ? SReg : `DATAWIDTH'bz;

reg     [`ADDR] addr_tbl[0:`HASHSIZE-1];
reg     [`DATA] data_tbl[0:`HASHSIZE-1];


initial begin : init_hashtbl
   integer i;
   for (i=0; i<`HASHSIZE; i=i+1)
      addr_tbl[i] = `ADDRWIDTH'bx;
end

initial begin : init_datatbl
   integer i;
   for (i=0; i<`HASHSIZE; i=i+1)
      data_tbl[i] = i;
end


function integer hashkey;
input   [`ADDR] addr;
input   i;
integer i;

   hashkey = (addr%`HASHSIZE + i*((addr/2)*2+1)) % `HASHSIZE;

endfunction /* hashkey */


function [`DATA] readmem;
input   [`ADDR] addr;
integer index, i;

begin
   index = hashkey(addr, 0);
   if (`dbg) $display(" read hashkey =         %d", index);

   if (addr_tbl[index] == addr)
      readmem = data_tbl[index];
   else begin : searchloop
      // double hashing
      for (i=1; i<`HASHSIZE; i=i+1) begin
         index = hashkey(addr, i);
         if (addr_tbl[index] == addr) begin
            readmem = data_tbl[index];
            disable searchloop;
         end
      end
      if (i == `HASHSIZE) begin
         $display(" Warning: read addr %h still uninitialized.", addr);
//         $stop;
      end
   end
end
endfunction /* readmem */


task writemem;
input   [`ADDR] addr;
input   [`DATA] word;
integer         index, i;

begin
   index = hashkey(addr, 0);
   if (`dbg) $display(" write hashkey =        %d", index);

   if (addr_tbl[index] == addr)     // found addr (initialized already)
      data_tbl[index] = word;
   else begin
      if (addr_tbl[index] === `ADDRWIDTH'bx) begin      // (addr still unused)
         addr_tbl[index] = addr;
         data_tbl[index] = word;
      end
      else begin : searchloop
         // double hashing
         for (i=1; i<`HASHSIZE; i=i+1) begin
            index = hashkey(addr, i);
            if (addr_tbl[index] !== `ADDRWIDTH'bx) begin
               if (addr_tbl[index] == addr) begin       // found addr
                  addr_tbl[index] = addr;
                  data_tbl[index] = word;
                  disable searchloop;
               end
            end
            else begin  // use next unused addr
               addr_tbl[index] = addr;
               data_tbl[index] = word;
               disable searchloop;
            end
         end
         if (i == `HASHSIZE) begin
            $display(" %m: Hashtable overflow!");
            $finish;
         end
      end
   end
end
endtask /* writemem */


always @(negedge Clk) begin
   if (Strobe) begin
      if (`dbg) $display($time, " hashmem> received strobe trigger.");
      @(posedge Clk) ;
      if (RWSelect == `READ) begin
         if (`dbg) $display($time, " hashmem> read request for address: %h", Addr);

         repeat (`READ_WAITCYCLES-1)
            @(posedge Clk) ;       // Just wait

         // Wait another clock cycle (Supply address to memory)
         @(posedge Clk) begin
            SReg = readmem(Addr);  // Supply Data to bus for one clock cycle
            if (`dbg) $display($time, " hashmem> supply data to bus:       %h", SReg);
         end

         @(posedge Clk)         // Done: Data still supplied but not valid!
            if (`dbg) $display($time, " hashmem> read request served.");
      end
      else if (RWSelect == `WRITE) begin
         if (`dbg) $display($time, " hashmem> writing data to memory...");
         writemem(Addr, Data);

         repeat (`WRITE_WAITCYCLES-1)
            @(posedge Clk) ;

         @(posedge Clk)
            if (`vbs) $display($time, " hashmem> stored at     %h: %h", Addr, Data);
      end
      else begin
         $display($time, " hashmem> Illegal RWSelect signal on bus: %b", RWSelect);
         $stop;
      end
   end
end

endmodule /* HashMem */
